import hashlib
import json

class Issue:

    def __init__(self, contract, function, pc, title, _type="Informational", description="", debug=""):

        self.title = title
        self.contract = contract
        self.function = function
        self.pc = pc
        self.description = description
        self.type = _type
        self.debug = debug
        self.filename = None
        self.code = None
        self.lineno = None

    def as_dict(self):

        issue = {'title': self.title, 'description':self.description, 'function': self.function, 'type': self.type, 'address': self.pc, 'debug': self.debug}

        if self.filename and self.lineno:
            issue['filename'] = self.filename
            issue['lineno'] = self.lineno

        if self.code:
            issue['code'] = self.code

        return issue

    def add_code_info(self, contract):
        if self.pc:
            codeinfo = contract.get_source_info(self.pc)
            self.filename = codeinfo.filename
            self.code = codeinfo.code
            self.lineno = codeinfo.lineno

class Report:

    def __init__(self, verbose=False):
        self.issues = {}
        self.verbose = verbose
        pass

    def append_issue(self, issue):
        m = hashlib.md5()
        m.update((issue.contract + str(issue.pc) + issue.title).encode('utf-8'))
        self.issues[m.digest()] = issue

    def as_text(self):
        text = ""

        for key, issue in self.issues.items():

            text += "==== " + issue.title + " ====\n"
            text += "Type: " + issue.type + "\n"

            if len(issue.contract):
                text += "Contract: " + issue.contract + "\n"
            else:
                text += "Contract: Unknown\n"

            text += "Function name: " + issue.function + "\n"
            text += "PC address: " + str(issue.pc) + "\n"

            text += issue.description + "\n--------------------\n"

            if issue.filename and issue.lineno:
                text += "In file: " + issue.filename + ":" + str(issue.lineno)

            if issue.code:
                text += "\n\n" + issue.code + "\n\n--------------------\n"

            if self.verbose and issue.debug:
                text += "\nDEBUGGING INFORMATION:\n\n" + issue.debug + "\n--------------------\n"

            text += "\n"

        return text

    def as_json(self):
        issues = []
        for key, issue in self.issues.items():
            issues.append(issue.as_dict())
        result = {'success': True, 'error': None, 'issues': issues}
        return json.dumps(result)

    def as_markdown(self):
        text = ""

        for key, issue in self.issues.items():

            if text == "":
                if (issue.filename):
                    text += "# Analysis results for " + issue.filename
               
            text += "\n\n## " + issue.title + "\n\n"
            text += "- Type: " + issue.type + "\n"

            if len(issue.contract):
                text += "- Contract: " + issue.contract + "\n"
            else:
                text += "- Contract: Unknown\n"

            text += "- Function name: `" + issue.function + "`\n"
            text += "- PC address: " + str(issue.pc) + "\n\n"

            text += "### Description\n\n" + issue.description

            if issue.filename and issue.lineno:
                text += "\nIn *%s:%d*\n" % (issue.filename, issue.lineno)

            if issue.code:
                text += "\n```\n" + issue.code + "\n```"

            if self.verbose and issue.debug:
                text += "\n\n### Debugging Information\n" + issue.debug

        return text



import os
import hashlib
import persistent.list
import transaction
from BTrees.OOBTree import BTree
import ZODB
from ZODB import FileStorage
from multiprocessing import Pool
import logging
from mythril.ether.ethcontract import ETHContract, InstanceList
from mythril import ether
import time

BLOCKS_PER_THREAD = 256
NUM_THREADS = 8


def get_persistent_storage(db_dir=None):
    if not db_dir:
        db_dir = os.path.join(os.path.expanduser('~'), ".mythril")

    if not os.path.exists(db_dir):
        os.makedirs(db_dir)

    db_path = os.path.join(db_dir, "contractstorage.fs")

    storage = FileStorage.FileStorage(db_path)
    db = ZODB.DB(storage)
    connection = db.open()
    storage_root = connection.root()

    try:
        contract_storage = storage_root['contractStorage']
    except KeyError:
        contract_storage = ContractStorage()
        storage_root['contractStorage'] = contract_storage

    return contract_storage, db

class SyncBlocks(object):
  '''
  Processes the block chunk
  '''

  def __init__(self, eth):
    self.eth = eth

  def __call__(self, startblock):
    '''
    Processesing method
    '''
    logging.info("SYNC_BLOCKS %d to %d" % (startblock, startblock + BLOCKS_PER_THREAD))

    contracts = {}

    for blockNum in range(startblock, startblock + BLOCKS_PER_THREAD):
        block = self.eth.eth_getBlockByNumber(blockNum)

        for tx in block['transactions']:

            if not tx['to']:

                receipt = self.eth.eth_getTransactionReceipt(tx['hash'])

                if receipt is not None:

                    contract_address = receipt['contractAddress']

                    contract_code = self.eth.eth_getCode(contract_address)
                    contract_balance = self.eth.eth_getBalance(contract_address)

                    if not contract_balance:
                        continue

                    ethcontract = ETHContract(contract_code, tx['input'])

                    m = hashlib.md5()
                    m.update(contract_code.encode('UTF-8'))
                    contract_hash = m.digest()

                    contracts[contract_hash] = {'ethcontract': ethcontract, 'address': contract_address, 'balance': contract_balance}

        blockNum -= 1

    return contracts

class ContractStorage(persistent.Persistent):

    def __init__(self):
        self.contracts = BTree()
        self.instance_lists = BTree()
        self.last_block = 0
        self.eth = None

    def get_contract_by_hash(self, contract_hash):
        return self.contracts[contract_hash]

    def initialize(self, eth):

        self.eth = eth

        if self.last_block:
            blockNum = self.last_block
            print("Resuming synchronization from block " + str(blockNum))
        else:

            blockNum = eth.eth_blockNumber()
            print("Starting synchronization from latest block: " + str(blockNum))

        processed = 0

        while (blockNum > 0):

            numbers = []

            for i in range(1, NUM_THREADS + 1):
                numbers.append(max(0, blockNum - (i * BLOCKS_PER_THREAD)))

            pool = Pool(NUM_THREADS, initargs=(self.eth))
            results = pool.map(SyncBlocks(self.eth), numbers)
            pool.close()
            pool.join()

            for result in results:
                for (contract_hash, data) in result.items():

                    try:
                        self.contracts[contract_hash]
                    except KeyError:
                        self.contracts[contract_hash] = data['ethcontract']
                        m = InstanceList()
                        self.instance_lists[contract_hash] = m

                    self.instance_lists[contract_hash].add(data['address'], data['balance'])

            blockNum -= NUM_THREADS * BLOCKS_PER_THREAD
            processed += NUM_THREADS * BLOCKS_PER_THREAD
            self.last_block = blockNum
            transaction.commit()

            cost_time = time.time() - ether.start_time
            print("%d blocks processed (in %d seconds), %d unique contracts in database, next block: %d" % (processed, cost_time, len(self.contracts), max(0, blockNum)))

        # If we've finished initializing the database, start over from the end of the chain if we want to initialize again
        self.last_block = 0
        print("Finished synchronization")

    def search(self, expression, callback_func):

        all_keys = list(self.contracts)

        for k in all_keys:

            if self.contracts[k].matches_expression(expression):
                m = self.instance_lists[k]

                callback_func(k.hex(), self.contracts[k], m.addresses, m.balances)



class MythrilBaseException(Exception): 
    pass


class CompilerError(MythrilBaseException): 
    pass


class UnsatError(MythrilBaseException): 
    pass


class NoContractFoundError(MythrilBaseException): 
    pass


class CriticalError(MythrilBaseException):
    pass




import json
import logging
import requests
from requests.adapters import HTTPAdapter
from requests.exceptions import ConnectionError as RequestsConnectionError
from .exceptions import (ConnectionError, BadStatusCodeError, BadJsonError, BadResponseError)
from .base_client import BaseClient

GETH_DEFAULT_RPC_PORT = 8545
ETH_DEFAULT_RPC_PORT = 8545
PARITY_DEFAULT_RPC_PORT = 8545
PYETHAPP_DEFAULT_RPC_PORT = 4000
MAX_RETRIES = 3
JSON_MEDIA_TYPE = 'application/json'

'''
This code is adapted from: https://github.com/ConsenSys/ethjsonrpc
'''
class EthJsonRpc(BaseClient):
    '''
    Ethereum JSON-RPC client class
    '''

    def __init__(self, host='localhost', port=GETH_DEFAULT_RPC_PORT, tls=False):
        self.host = host
        self.port = port
        self.tls = tls
        self.session = requests.Session()
        self.session.mount(self.host, HTTPAdapter(max_retries=MAX_RETRIES))

    def _call(self, method, params=None, _id=1):

        params = params or []
        data = {
            'jsonrpc': '2.0',
            'method': method,
            'params': params,
            'id': _id,
        }
        scheme = 'http'
        if self.tls:
            scheme += 's'
        url = '{}://{}:{}'.format(scheme, self.host, self.port)
        headers = {'Content-Type': JSON_MEDIA_TYPE}
        logging.debug("rpc send: %s" % json.dumps(data))
        try:
            r = self.session.post(url, headers=headers, data=json.dumps(data))
        except RequestsConnectionError:
            raise ConnectionError
        if r.status_code / 100 != 2:
            raise BadStatusCodeError(r.status_code)
        try:
            response = r.json()
            logging.debug("rpc response: %s" % response)
        except ValueError:
            raise BadJsonError(r.text)
        try:
            return response['result']
        except KeyError:
            raise BadResponseError(response)

    def close(self):
        self.session.close()



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
"""mythril.py: Bug hunting on the Ethereum blockchain

   http://www.github.com/b-mueller/mythril
"""

import logging
import json
import os
import re

from ethereum import utils
from solc.exceptions import SolcError
import solc

from mythril.ether import util
from mythril.ether.contractstorage import get_persistent_storage
from mythril.ether.ethcontract import ETHContract
from mythril.ether.soliditycontract import SolidityContract
from mythril.rpc.client import EthJsonRpc
from mythril.ipc.client import EthIpc
from mythril.rpc.exceptions import ConnectionError
from mythril.support import signatures
from mythril.support.truffle import analyze_truffle_project
from mythril.support.loader import DynLoader
from mythril.exceptions import CompilerError, NoContractFoundError, CriticalError
from mythril.analysis.symbolic import SymExecWrapper
from mythril.analysis.callgraph import generate_graph
from mythril.analysis.traceexplore import get_serializable_statespace
from mythril.analysis.security import fire_lasers
from mythril.analysis.report import Report
from mythril.leveldb.client import EthLevelDB


# logging.basicConfig(level=logging.DEBUG)

class Mythril(object):
    """
    Mythril main interface class. 

    1. create mythril object
    2. set rpc or leveldb interface if needed
    3. load contracts (from solidity, bytecode, address)
    4. fire_lasers

    Example:
        mythril = Mythril()
        mythril.set_db_rpc_infura()

        # (optional) other db adapters
        mythril.set_db_rpc(args)
        mythril.set_db_ipc()
        mythril.set_db_rpc_localhost()

        # (optional) other func
        mythril.analyze_truffle_project(args)
        mythril.search_db(args)
        mythril.init_db()

        # load contract
        mythril.load_from_bytecode(bytecode)
        mythril.load_from_address(address)
        mythril.load_from_solidity(solidity_file)

        # analyze
        print(mythril.fire_lasers(args).as_text())

        # (optional) graph
        for contract in mythril.contracts:
            print(mythril.graph_html(args))  # prints html or save it to file
        
        # (optional) other funcs
        mythril.dump_statespaces(args)
        mythril.disassemble(contract)
        mythril.get_state_variable_from_storage(args)

    """


    def __init__(self, solv=None,
                 solc_args=None, dynld=False):

        self.solv = solv
        self.solc_args = solc_args
        self.dynld = dynld

        self.mythril_dir = self._init_mythril_dir()
        self.signatures_file, self.sigs = self._init_signatures()
        self.solc_binary = self._init_solc_binary(solv)

        self.eth = None
        self.ethDb = None
        self.dbtype = None  # track type of db (rpc,ipc,leveldb) used

        self.contracts = []  # loaded contracts

    def _init_mythril_dir(self):
        try:
            mythril_dir = os.environ['MYTHRIL_DIR']
        except KeyError:
            mythril_dir = os.path.join(os.path.expanduser('~'), ".mythril")

            # Initialize data directory and signature database

        if not os.path.exists(mythril_dir):
            logging.info("Creating mythril data directory")
            os.mkdir(mythril_dir)
        return mythril_dir

    def _init_signatures(self):

        # If no function signature file exists, create it. Function signatures from Solidity source code are added automatically.

        signatures_file = os.path.join(self.mythril_dir, 'signatures.json')

        sigs = {}
        if not os.path.exists(signatures_file):
            logging.info("No signature database found. Creating empty database: " + signatures_file + "\n" +
                         "Consider replacing it with the pre-initialized database at https://raw.githubusercontent.com/ConsenSys/mythril/master/signatures.json")
            with open(signatures_file, 'a') as f:
                json.dump({}, f)

        with open(signatures_file) as f:
            try:
                sigs = json.load(f)
            except json.JSONDecodeError as e:
                raise CriticalError("Invalid JSON in signatures file " + signatures_file + "\n" + str(e))
        return signatures_file, sigs

    def _update_signatures(self, jsonsigs):
        # Save updated function signatures
        with open(self.signatures_file, 'w') as f:
            json.dump(jsonsigs, f)

        self.sigs = jsonsigs

    def analyze_truffle_project(self, *args, **kwargs):
        return analyze_truffle_project(*args, **kwargs)  # just passthru for now

    def _init_solc_binary(self, version):
        # Figure out solc binary and version
        # Only proper versions are supported. No nightlies, commits etc (such as available in remix)

        if version:
            # tried converting input to semver, seemed not necessary so just slicing for now
            if version == str(solc.main.get_solc_version())[:6]:
                logging.info('Given version matches installed version')
                try:
                    solc_binary = os.environ['SOLC']
                except KeyError:
                    solc_binary = 'solc'
            else:
                if util.solc_exists(version):
                    logging.info('Given version is already installed')
                else:
                    try:
                        solc.install_solc('v' + version)
                    except SolcError:
                        raise CriticalError("There was an error when trying to install the specified solc version")

                solc_binary = os.path.join(os.environ['HOME'], ".py-solc/solc-v" + version, "bin/solc")
                logging.info("Setting the compiler to " + str(solc_binary))
        else:
            try:
                solc_binary = os.environ['SOLC']
            except KeyError:
                solc_binary = 'solc'
        return solc_binary

    def set_db_leveldb(self, leveldb):
        self.ethDb = EthLevelDB(leveldb)
        self.eth = self.ethDb
        self.dbtype = "leveldb"
        return self.eth

    def set_db_rpc_infura(self):
        self.eth = EthJsonRpc('mainnet.infura.io', 443, True)
        logging.info("Using INFURA for RPC queries")
        self.dbtype = "rpc"

    def set_db_rpc(self, rpc=None, rpctls=False):
        if rpc == 'ganache':
            rpcconfig = ('localhost', 7545, False)
        else:
            m = re.match(r'infura-(.*)', rpc)
            if m and m.group(1) in ['mainnet', 'rinkeby', 'kovan', 'ropsten']:
                rpcconfig = (m.group(1) + '.infura.io', 443, True)
            else:
                try:
                    host, port = rpc.split(":")
                    rpcconfig = (host, int(port), rpctls)
                except ValueError:
                    raise CriticalError("Invalid RPC argument, use 'ganache', 'infura-[network]' or 'HOST:PORT'")

        if rpcconfig:
            self.eth = EthJsonRpc(rpcconfig[0], int(rpcconfig[1]), rpcconfig[2])
            self.dbtype = "rpc"
            logging.info("Using RPC settings: %s" % str(rpcconfig))
        else:
            raise CriticalError("Invalid RPC settings, check help for details.")

    def set_db_ipc(self):
        try:
            self.eth = EthIpc()
            self.dbtype = "ipc"
        except Exception as e:
            raise CriticalError(
                "IPC initialization failed. Please verify that your local Ethereum node is running, or use the -i flag to connect to INFURA. \n" + str(
                    e))

    def set_db_rpc_localhost(self):
        self.eth = EthJsonRpc('localhost', 8545)
        self.dbtype = "rpc"
        logging.info("Using default RPC settings: http://localhost:8545")

    def search_db(self, search):

        def search_callback(code_hash, code, addresses, balances):
            print("Matched contract with code hash " + code_hash)
            for i in range(0, len(addresses)):
                print("Address: " + addresses[i] + ", balance: " + str(balances[i]))

        contract_storage, _ = get_persistent_storage(self.mythril_dir)
        try:
            if self.dbtype=="leveldb":
                contract_storage.search(search, search_callback)
            else:
                self.ethDB.search(search, search_callback)
        except SyntaxError:
            raise CriticalError("Syntax error in search expression.")

    def init_db(self):
        contract_storage, _ = get_persistent_storage(self.mythril_dir)
        try:
            contract_storage.initialize(self.eth)
        except FileNotFoundError as e:
             raise CriticalError("Error syncing database over IPC: " + str(e))
        except ConnectionError as e:
            raise CriticalError("Could not connect to RPC server. Make sure that your node is running and that RPC parameters are set correctly.")

    def load_from_bytecode(self, code):
        address = util.get_indexed_address(0)
        self.contracts.append(ETHContract(code, name="MAIN"))
        return address, self.contracts[-1]  # return address and contract object

    def load_from_address(self, address):
        if not re.match(r'0x[a-fA-F0-9]{40}', address):
             raise CriticalError("Invalid contract address. Expected format is '0x...'.")

        try:
            code = self.eth.eth_getCode(address)
        except FileNotFoundError as e:
             raise CriticalError("IPC error: " + str(e))
        except ConnectionError as e:
            raise CriticalError("Could not connect to RPC server. Make sure that your node is running and that RPC parameters are set correctly.")
        except Exception as e:
             raise CriticalError("IPC / RPC error: " + str(e))
        else:
            if code == "0x" or code == "0x0":
                 raise CriticalError("Received an empty response from eth_getCode. Check the contract address and verify that you are on the correct chain.")
            else:
                self.contracts.append(ETHContract(code, name=address))
        return address, self.contracts[-1]  # return address and contract object

    def load_from_solidity(self, solidity_files):
        """
        UPDATES self.sigs!
        :param solidity_files:
        :return:
        """
        address = util.get_indexed_address(0)
        contracts = []
        for file in solidity_files:
            if ":" in file:
                file, contract_name = file.split(":")
            else:
                contract_name = None

            file = os.path.expanduser(file)

            try:
                signatures.add_signatures_from_file(file, self.sigs)
                contract = SolidityContract(file, contract_name, solc_args=self.solc_args)
                logging.info("Analyzing contract %s:%s" % (file, contract.name))
            except FileNotFoundError:
                 raise CriticalError("Input file not found: " + file)
            except CompilerError as e:
                 raise CriticalError(e)
            except NoContractFoundError:
                logging.info("The file " + file + " does not contain a compilable contract.")
            else:
                self.contracts.append(contract)
                contracts.append(contract)

        self._update_signatures(self.sigs)
        return address, contracts

    def dump_statespaces(self, contracts=None, address=None, max_depth=12):
        statespaces = []

        for contract in (contracts or self.contracts):
            sym = SymExecWrapper(contract, address,
                                 dynloader=DynLoader(self.eth) if self.dynld else None,
                                 max_depth=max_depth)
            statespaces.append((contract, get_serializable_statespace(sym)))

        return statespaces

    def graph_html(self, contract, address, max_depth=12, enable_physics=False, phrackify=False):
        sym = SymExecWrapper(contract, address,
                             dynloader=DynLoader(self.eth) if self.dynld else None,
                             max_depth=max_depth)
        return generate_graph(sym, physics=enable_physics, phrackify=phrackify)

    def fire_lasers(self, contracts=None, address=None,
                    modules=None,
                    verbose_report=False, max_depth=12):

        all_issues = []
        for contract in (contracts or self.contracts):

            sym = SymExecWrapper(contract, address,
                                 dynloader=DynLoader(self.eth) if self.dynld else None,
                                 max_depth=max_depth)

            issues = fire_lasers(sym, modules)

            if type(contract) == SolidityContract:
                for issue in issues:
                    issue.add_code_info(contract)

                all_issues += issues

        # Finally, output the results
        report = Report(verbose_report)
        for issue in all_issues:
            report.append_issue(issue)

        return report

    def get_state_variable_from_storage(self, address, params=[]):
        (position, length, mappings) = (0, 1, [])
        try:
            if params[0] == "mapping":
                if len(params) < 3:
                     raise CriticalError("Invalid number of parameters.")
                position = int(params[1])
                position_formatted = utils.zpad(utils.int_to_big_endian(position), 32)
                for i in range(2, len(params)):
                    key = bytes(params[i], 'utf8')
                    key_formatted = utils.rzpad(key, 32)
                    mappings.append(int.from_bytes(utils.sha3(key_formatted + position_formatted), byteorder='big'))

                length = len(mappings)
                if length == 1:
                    position = mappings[0]

            else:
                if len(params) >= 4:
                     raise CriticalError("Invalid number of parameters.")

                if len(params) >= 1:
                    position = int(params[0])
                if len(params) >= 2:
                    length = int(params[1])
                if len(params) == 3 and params[2] == "array":
                    position_formatted = utils.zpad(utils.int_to_big_endian(position), 32)
                    position = int.from_bytes(utils.sha3(position_formatted), byteorder='big')

        except ValueError:
             raise CriticalError("Invalid storage index. Please provide a numeric value.")

        outtxt = []

        try:
            if length == 1:
                outtxt.append("{}: {}".format(position, self.eth.eth_getStorageAt(address, position)))
            else:
                if len(mappings) > 0:
                    for i in range(0, len(mappings)):
                        position = mappings[i]
                        outtxt.append("{}: {}".format(hex(position), self.eth.eth_getStorageAt(address, position)))
                else:
                    for i in range(position, position + length):
                        outtxt.append("{}: {}".format(hex(i), self.eth.eth_getStorageAt(address, i)))
        except FileNotFoundError as e:
             raise CriticalError("IPC error: " + str(e))
        except ConnectionError as e:
            raise CriticalError("Could not connect to RPC server. Make sure that your node is running and that RPC parameters are set correctly.")
        return '\n'.join(outtxt)

    def disassemble(self, contract):
        return contract.get_easm()

    @staticmethod
    def hash_for_function_signature(sig):
        return "0x%s" % utils.sha3(sig)[:4].hex()



from mythril.analysis.report import Issue
from laser.ethereum.svm import NodeFlags
import logging
import re


'''
MODULE DESCRIPTION:

Test whether CALL return value is checked.

For direct calls, the Solidity compiler auto-generates this check. E.g.:

    Alice c = Alice(address);
    c.ping(42);

Here the CALL will be followed by IZSERO(retval), if retval = ZERO then state is reverted.

For low-level-calls this check is omitted. E.g.:

    c.call.value(0)(bytes4(sha3("ping(uint256)")),1);

'''


def execute(statespace):

    logging.debug("Executing module: UNCHECKED_RETVAL")

    issues = []

    for k in statespace.nodes:

        node = statespace.nodes[k]

        if NodeFlags.CALL_RETURN in node.flags:

            retval_checked = False

            for state in node.states:

                instr = state.get_current_instruction()

                if (instr['opcode'] == 'ISZERO' and re.search(r'retval', str(state.mstate.stack[-1]))):
                    retval_checked = True
                    break

            if not retval_checked:

                address = state.get_current_instruction()['address']
                issue = Issue(node.contract_name, node.function_name, address, "Unchecked CALL return value")

                issue.description = \
                    "The return value of an external call is not checked. Note that execution continue even if the called contract throws."

                issues.append(issue)

        else:

            nStates = len(node.states)

            for idx in range(0, nStates - 1):  # Ignore CALLs at last position in a node

                state = node.states[idx]
                instr = state.get_current_instruction()

                if (instr['opcode'] == 'CALL'):

                    retval_checked = False

                    for _idx in range(idx, idx + 10):

                        try:
                            _state = node.states[_idx]
                            _instr = _state.get_current_instruction()

                            if (_instr['opcode'] == 'ISZERO' and re.search(r'retval', str(_state .mstate.stack[-1]))):
                                retval_checked = True
                                break

                        except IndexError:
                            break

                    if not retval_checked:

                        address = instr['address']
                        issue = Issue(node.contract_name, node.function_name, address, "Unchecked CALL return value")

                        issue.description = \
                            "The return value of an external call is not checked. Note that execution continue even if the called contract throws."

                        issues.append(issue)

    return issues



from z3 import *
from mythril.analysis import solver
from mythril.analysis.ops import *
from mythril.analysis.report import Issue
from mythril.exceptions import UnsatError
import re
import copy
import logging

'''
MODULE DESCRIPTION:

Check for integer underflows.
For every SUB instruction, check if there's a possible state where op1 > op0.
For every ADD, MUL instruction, check if there's a possible state where op1 + op0 > 2^32 - 1
'''

def execute(statespace):
    """
    Executes analysis module for integer underflow and integer overflow
    :param statespace: Statespace to analyse
    :return: Found issues
    """
    logging.debug("Executing module: INTEGER")

    issues = []

    for k in statespace.nodes:
        node = statespace.nodes[k]

        for state in node.states:
            logging.debug("Checking for integer underflow")
            issues += _check_integer_underflow(statespace, state, node)
            logging.debug("Checking for integer overflow")
            issues += _check_integer_overflow(statespace, state, node)

    return issues


def _check_integer_overflow(statespace, state, node):
    """
    Checks for integer overflow
    :param statespace: statespace that is being examined
    :param state: state from node to examine
    :param node: node to examine
    :return: found issue
    """
    issues = []

    # Check the instruction
    instruction = state.get_current_instruction()
    if instruction['opcode'] not in ("ADD", "MUL"):
        return issues

    # Formulate overflow constraints
    stack = state.mstate.stack
    op0, op1 = stack[-1], stack[-2]

    # An integer overflow is possible if op0 + op1 or op0 * op1 > MAX_UINT
    # Do a type check
    allowed_types = [int, BitVecRef, BitVecNumRef]
    if not (type(op0) in allowed_types and type(op1) in allowed_types):
        return issues

    # Change ints to BitVec
    if type(op0) is int:
        op0 = BitVecVal(op0, 256)
    if type(op1) is int:
        op1 = BitVecVal(op1, 256)

    # Formulate expression
    if instruction['opcode'] == "ADD":
        expr = op0 + op1
    else:
        expr = op1 * op0

    # Check satisfiable
    constraint = Or(ULT(expr, op0), ULT(expr, op1))
    model = _try_constraints(node.constraints, [constraint])

    if model is None:
        logging.debug("[INTEGER_OVERFLOW] no model found")
        return issues

    if not _verify_integer_overflow(statespace, node, expr, state, model, constraint, op0, op1):
        return issues

    # Build issue
    issue = Issue(node.contract_name, node.function_name, instruction['address'], "Integer Overflow ", "Warning")

    issue.description = "A possible integer overflow exists in the function `{}`.\n" \
                        "The addition or multiplication may result in a value higher than the maximum representable integer.".format(
        node.function_name)
    issue.debug = solver.pretty_print_model(model)
    issues.append(issue)

    return issues


def _verify_integer_overflow(statespace, node, expr, state, model, constraint, op0, op1):
    """ Verifies existence of integer overflow """
    # If we get to this point then there has been an integer overflow
    # Find out if the overflowed value is actually used
    interesting_usages = _search_children(statespace, node, expr, constraint=[constraint], index=node.states.index(state))

    # Stop if it isn't
    if len(interesting_usages) == 0:
        return False

    model0, model1 = None, None

    if type(op0) is not int:
        op0_value = int(str(model.eval(op0, model_completion=True)))
        model0 = _try_constraints(node.constraints, [constraint, op0 != op0_value])

    if type(op1) is not int:
        op1_value = int(str(model.eval(op1, model_completion=True)))
        model1 = _try_constraints(node.constraints, [constraint, op1 != op1_value])

    if model0 is None and model1 is None:
        return False

    return True


def _try_constraints(constraints, new_constraints):
    """
    Tries new constraints
    :return Model if satisfiable otherwise None
    """
    _constraints = copy.deepcopy(constraints)
    for constraint in new_constraints:
        _constraints.append(copy.deepcopy(constraint))
    try:
        model = solver.get_model(_constraints)
        return model
    except UnsatError:
        return None


def _check_integer_underflow(statespace, state, node):
    """
    Checks for integer underflow
    :param state: state from node to examine
    :param node: node to examine
    :return: found issue
    """
    issues = []
    instruction = state.get_current_instruction()

    if instruction['opcode'] == "SUB":

        stack = state.mstate.stack

        op0 = stack[-1]
        op1 = stack[-2]

        constraints = copy.deepcopy(node.constraints)

        # Filter for patterns that indicate benign underflows

        # Pattern 1: (96 + calldatasize_MAIN) - (96), where (96 + calldatasize_MAIN) would underflow if calldatasize is very large.
        # Pattern 2: (256*If(1 & storage_0 == 0, 1, 0)) - 1, this would underlow if storage_0 = 0
        if type(op0) == int and type(op1) == int:
            return []
        if re.search(r'calldatasize_', str(op0)):
            return []
        if re.search(r'256\*.*If\(1', str(op0), re.DOTALL) or re.search(r'256\*.*If\(1', str(op1), re.DOTALL):
            return []
        if re.search(r'32 \+.*calldata', str(op0), re.DOTALL) or re.search(r'32 \+.*calldata', str(op1), re.DOTALL):
            return []

        logging.debug("[INTEGER_UNDERFLOW] Checking SUB {0}, {1} at address {2}".format(str(op0), str(op1),
                                                                                        str(instruction['address'])))
        allowed_types = [int, BitVecRef, BitVecNumRef]

        if type(op0) in allowed_types and type(op1) in allowed_types:
            constraints.append(UGT(op1, op0))

            try:

                model = solver.get_model(constraints)

                # If we get to this point then there has been an integer overflow
                # Find out if the overflowed value is actually used
                interesting_usages = _search_children(statespace, node, (op0 - op1), index=node.states.index(state))
                logging.debug(interesting_usages)

                # Stop if it isn't
                if len(interesting_usages) == 0:
                    return issues

                issue = Issue(node.contract_name, node.function_name, instruction['address'], "Integer Underflow",
                              "Warning")

                issue.description = "A possible integer underflow exists in the function `" + node.function_name + "`.\n" \
                                                                                                                   "The subtraction may result in a value < 0."

                issue.debug = solver.pretty_print_model(model)
                issues.append(issue)

            except UnsatError:
                logging.debug("[INTEGER_UNDERFLOW] no model found")
    return issues


def _check_usage(state, expression):
    """Delegates checks to _check_{instruction_name}()"""
    opcode = state.get_current_instruction()['opcode']

    if opcode == 'JUMPI':
        if _check_jumpi(state, expression):
            return [state]
    elif opcode == 'SSTORE':
        if _check_sstore(state, expression):
            return [state]
    return []


def _check_taint(statement, expression):
    """Checks if statement is influenced by tainted expression"""
    _expression, _statement = str(expression).replace(' ', ''), str(statement).replace(' ', '')
    found = _expression in _statement

    if found:
        i = _statement.index(_expression)
        char = _statement[i - 1]
        if char == '_':
            return False
    return found


def _check_jumpi(state, expression):
    """ Check if conditional jump is dependent on the result of expression"""
    logging.debug(state.get_current_instruction()['opcode'])
    assert state.get_current_instruction()['opcode'] == 'JUMPI'
    condition = state.mstate.stack[-2]
    return _check_taint(condition, expression)


def _check_sstore(state, expression):
    """ Check if store operation is dependent on the result of expression"""
    logging.debug(state.get_current_instruction()['opcode'])
    assert state.get_current_instruction()['opcode'] == 'SSTORE'
    value = state.mstate.stack[-2]
    return _check_taint(value, expression)


def _search_children(statespace, node, expression, constraint=[], index=0, depth=0, max_depth=64):
    """
    Checks the statespace for children states, with JUMPI or SSTORE instuctions,
    for dependency on expression
    :param statespace: The statespace to explore
    :param node: Current node to explore from
    :param expression: Expression to look for
    :param index: Current state index node.states[index] == current_state
    :param depth: Current depth level
    :param max_depth: Max depth to explore
    :return: List of states that match the opcodes and are dependent on expression
    """
    logging.debug("SEARCHING NODE for usage of an overflowed variable %d", node.uid)

    results = []

    if depth >= max_depth:
        return []

    # Explore current node from index
    for j in range(index, len(node.states)):
        current_state = node.states[j]
        current_instruction = current_state.get_current_instruction()
        if current_instruction['opcode'] in ('JUMPI', 'SSTORE'):
            element = _check_usage(current_state, expression)
            if len(element) < 1:
                continue
            # if _check_requires(element[0], node, statespace, constraint):
            #     pass
            results += element

    # Recursively search children
    children = \
        [
            statespace.nodes[edge.node_to]
            for edge in statespace.edges
            if edge.node_from == node.uid
            # and _try_constraints(statespace.nodes[edge.node_to].constraints, constraint) is not None
        ]

    for child in children:
        results += _search_children(statespace, child, expression, depth=depth + 1, max_depth=max_depth)

    return results


def _check_requires(state, node, statespace, constraint):
    """Checks if usage of overflowed statement results in a revert statement"""
    instruction = state.get_current_instruction()
    if instruction['opcode'] is not "JUMPI":
        return False
    children = [
            statespace.nodes[edge.node_to]
            for edge in statespace.edges
            if edge.node_from == node.uid
        ]

    for child in children:
        opcodes = [s.get_current_instruction()['opcode'] for s in child.states]
        if "REVERT" in opcodes or "ASSERT_FAIL" in opcodes:
            return True
    # I added the following case, bc of false positives if the max depth is not high enough
    if len(children) == 0:
        return True
    return False



from mythril.analysis.report import Issue
from mythril.exceptions import UnsatError
from mythril.analysis import solver
import logging


'''
MODULE DESCRIPTION:

Checks whether any exception states are reachable.

'''


def execute(statespace):

    logging.debug("Executing module: EXCEPTIONS")

    issues = []

    for k in statespace.nodes:
        node = statespace.nodes[k]

        for state in node.states:

            instruction = state.get_current_instruction()

            if(instruction['opcode'] == "ASSERT_FAIL"):

                try:
                    model = solver.get_model(node.constraints)
                    address = state.get_current_instruction()['address']

                    description = "A reachable exception (opcode 0xfe) has been detected. This can be caused by type errors, division by zero, out-of-bounds array access, or assert violations. "
                    description += "This is acceptable in most situations. Note however that `assert()` should only be used to check invariants. Use `require()` for regular input checking. "

                    debug = "The exception is triggered under the following conditions:\n\n"

                    debug += solver.pretty_print_model(model)

                    issues.append(Issue(node.contract_name, node.function_name, address, "Exception state", "Informational", description, debug))

                except UnsatError:
                    logging.debug("[EXCEPTIONS] no model found")

    return issues



from z3 import *
from enum import Enum
from laser.ethereum import helper


class VarType(Enum):
    SYMBOLIC = 1
    CONCRETE = 2


class Variable:

    def __init__(self, val, _type):
        self.val = val
        self.type = _type

    def __str__(self):
        return str(self.val)


def get_variable(i):
    try:
        return Variable(helper.get_concrete_int(i), VarType.CONCRETE)
    except AttributeError:
        return Variable(simplify(i), VarType.SYMBOLIC)


class Op:

    def __init__(self, node, state, state_index):
        self.node = node
        self.state = state
        self.state_index = state_index


class Call(Op):

    def __init__(self, node, state, state_index, _type, to, gas, value=Variable(0, VarType.CONCRETE), data=None):

        super().__init__(node, state, state_index)
        self.to = to
        self.gas = gas
        self.type = _type
        self.value = value
        self.data = data


class SStore(Op):

    def __init__(self, node, state, state_index, value):
        super().__init__(node, state, state_index)
        self.value = value



from z3 import Solver, simplify, sat
from mythril.exceptions import UnsatError


def get_model(constraints):
    s = Solver()
    s.set("timeout", 10000)

    for constraint in constraints:
        s.add(constraint)

    if (s.check() == sat):

        return s.model()

    else:
        raise UnsatError


def pretty_print_model(model):

    ret = ""

    for d in model.decls():

        try:
            condition = "0x%x" % model[d].as_long()
        except:
            condition = str(simplify(model[d]))

        ret += ("%s: %s\n" % (d.name(), condition))

    return ret



import os
import sys

from .constants import BLOCK_TAGS


def hex_to_dec(x):
    '''
    Convert hex to decimal
    '''
    return int(x, 16)

def to_bytes(text):
    return text.encode('utf-8')

def to_text(byte_array):
    return byte_array.decode('utf-8')

def clean_hex(d):
    '''
    Convert decimal to hex and remove the "L" suffix that is appended to large
    numbers
    '''
    return hex(d).rstrip('L')

def validate_block(block):
    if isinstance(block, str):
        if block not in BLOCK_TAGS:
            raise ValueError('invalid block tag')
    if isinstance(block, int):
        block = hex(block)
    return block


def wei_to_ether(wei):
    '''
    Convert wei to ether
    '''
    return 1.0 * wei / 10**18


def ether_to_wei(ether):
    '''
    Convert ether to wei
    '''
    return ether * 10**18


def get_default_ipc_path(testnet=False):
    if testnet:
        testnet = "testnet"
    else:
        testnet = ""

    if sys.platform == 'darwin':
        ipc_path = os.path.expanduser(os.path.join("~", "Library", "Ethereum", testnet, "geth.ipc"))
        if os.path.exists(ipc_path):
            return ipc_path

        ipc_path = os.path.expanduser(os.path.join("~", "Library", "Application Support", "io.parity.ethereum", "jsonrpc.ipc"))
        if os.path.exists(ipc_path):
            return ipc_path

    elif sys.platform.startswith('linux'):
        ipc_path = os.path.expanduser(os.path.join("~", ".ethereum", testnet, "geth.ipc"))
        if os.path.exists(ipc_path):
            return ipc_path

        ipc_path = os.path.expanduser(os.path.join("~", ".local", "share", "io.parity.ethereum", "jsonrpc.ipc"))
        if os.path.exists(ipc_path):
            return ipc_path

    elif sys.platform == 'win32':
        ipc_path = os.path.join("\\\\", ".", "pipe", "geth.ipc")
        if os.path.exists(ipc_path):
            return ipc_path

        ipc_path = os.path.join("\\\\", ".", "pipe", "jsonrpc.ipc")
        if os.path.exists(ipc_path):
            return ipc_path

    else:
        raise ValueError(
            "Unsupported platform '{0}'.  Only darwin/linux2/win32 are "
            "supported.  You must specify the ipc_path".format(sys.platform)
        )


import plyvel
import binascii
import rlp
import hashlib
from ethereum import utils
from ethereum.block import BlockHeader, Block
from mythril.leveldb.state import State, Account
from mythril.leveldb.eth_db import ETH_DB
from mythril.ether.ethcontract import ETHContract, InstanceList

# Per https://github.com/ethereum/go-ethereum/blob/master/core/database_util.go
# prefixes and suffixes for keys in geth
headerPrefix = b'h'     # headerPrefix + num (uint64 big endian) + hash -> header
bodyPrefix = b'b'       # bodyPrefix + num (uint64 big endian) + hash -> block body
numSuffix = b'n'        # headerPrefix + num (uint64 big endian) + numSuffix -> hash
blockHashPrefix = b'H'  # blockHashPrefix + hash -> num (uint64 big endian)
# known geth keys
headHeaderKey = b'LastBlock' # head (latest) header hash

def _formatBlockNumber(number):
    '''
    formats block number to uint64 big endian
    '''
    return utils.zpad(utils.int_to_big_endian(number), 8)

def _encode_hex(v):
    '''
    encodes hash as hex
    '''
    return '0x' + utils.encode_hex(v)

class EthLevelDB(object):
    '''
    Go-Ethereum LevelDB client class
    '''

    def __init__(self, path):
        self.path = path
        self.db = ETH_DB(path)
        self.headBlockHeader = None
        self.headState = None
        self.all_contracts = None
        self.active_contracts = None
        self.instance_lists = None

    def get_all_contracts(self):
        '''
        get all contracts
        '''
        if not self.all_contracts:
            self.all_contracts = []
            self.active_contracts = []
            self.instance_lists = []
            state = self._get_head_state()
            accounts = state.get_all_accounts()

            for a in accounts:
                if a.code is not None:
                    code = _encode_hex(a.code)
                    md5 = hashlib.md5()
                    md5.update(code.encode('UTF-8'))
                    contract_hash = md5.digest()
                    contract = ETHContract(code, name=contract_hash.hex())
                    self.all_contracts.append(contract)

                    if a.balance != 0:
                        md5 = InstanceList()
                        md5.add(_encode_hex(a.address), a.balance)
                        self.instance_lists.append(md5)
                        self.active_contracts.append(contract)

        return self.all_contracts

    def get_active_contracts(self):
        '''
        get all contracts with non-zero balance
        '''
        if not self.active_contracts:
            self.get_all_contracts() # optimized
        return self.active_contracts

    def search(self, expression, callback_func):
        '''
        searches through non-zero balance contracts
        '''
        contracts = self.get_active_contracts()
        for i in range(0, len(contracts)):
            if contracts[i].matches_expression(expression):
                m = self.instance_lists[i]
                callback_func(contracts[i].name, contracts[i], m.addresses, m.balances)

    def eth_getBlockHeaderByNumber(self, number):
        '''
        gets block header by block number
        '''
        hash = self._get_block_hash(number)
        blockNumber = _formatBlockNumber(number)
        return self._get_block_header(hash, blockNumber)

    def eth_getBlockByNumber(self, number):
        '''
        gets block body by block number
        '''
        blockHash = self._get_block_hash(number)
        blockNumber = _formatBlockNumber(number)
        bodyKey = bodyPrefix + blockNumber + blockHash
        blockData = self.db.get(bodyKey)
        body = rlp.decode(blockData, sedes=Block)
        return body

    def eth_getCode(self, address):
        '''
        gets account code
        '''
        account = self._get_account(address)
        return _encode_hex(account.code)

    def eth_getBalance(self, address):
        '''
        gets account balance
        '''
        account = self._get_account(address)
        return account.balance

    def eth_getStorageAt(self, address, position):
        '''
        gets account storage data at position
        '''
        account = self._get_account(address)
        return _encode_hex(utils.zpad(utils.encode_int(account.get_storage_data(position)), 32))

    def _get_head_state(self):
        '''
        gets head state
        '''
        if not self.headState:
            root = self._get_head_block().state_root
            self.headState = State(self.db, root)
        return self.headState

    def _get_account(self, address):
        '''
        gets account by address
        '''
        state = self._get_head_state()
        accountAddress = binascii.a2b_hex(utils.remove_0x_head(address))
        return state.get_and_cache_account(accountAddress)

    def _get_block_hash(self, number):
        '''
        gets block hash by block number
        '''
        num = _formatBlockNumber(number)
        hashKey = headerPrefix + num + numSuffix
        return self.db.get(hashKey)

    def _get_head_block(self):
        '''
        gets head block header
        '''
        if not self.headBlockHeader:
            hash = self.db.get(headHeaderKey)
            num = self._get_block_number(hash)
            self.headBlockHeader = self._get_block_header(hash, num)
            # find header with valid state
            while not self.db.get(self.headBlockHeader.state_root) and self.headBlockHeader.prevhash is not None:
                hash = self.headBlockHeader.prevhash
                num = self._get_block_number(hash)
                self.headBlockHeader = self._get_block_header(hash, num)
            
        return self.headBlockHeader

    def _get_block_number(self, hash):
        '''
        gets block number by hash
        '''
        numberKey = blockHashPrefix + hash
        return self.db.get(numberKey)

    def _get_block_header(self, hash, num):
        '''
        get block header by block header hash & number
        '''
        headerKey = headerPrefix + num + hash
        blockHeaderData = self.db.get(headerKey)
        header = rlp.decode(blockHeaderData, sedes=BlockHeader)
        return header


import json
import socket
import logging
from mythril.rpc.base_client import BaseClient
from .utils import (get_default_ipc_path, to_text, to_bytes)
import threading

try:
    from json import JSONDecodeError
except ImportError:
    JSONDecodeError = ValueError

IPC_PATH = None

'''
This code is mostly adapted from:
- https://github.com/ConsenSys/ethjsonrpc
- https://github.com/pipermerriam/ethereum-ipc-client
- https://github.com/ethereum/web3.py
'''

# use thread local to store socket which will be reused just in owner thread
THREAD_LOCAL = threading.local()

class EthIpc(BaseClient):

    def __init__(self, ipc_path=IPC_PATH, testnet=False):
        if ipc_path is None:
            ipc_path = get_default_ipc_path(testnet)
        self.ipc_path = ipc_path

    def connect_if_not_yet(self):
        if not hasattr(THREAD_LOCAL, "socket"):
            THREAD_LOCAL.socket = self.get_socket()

    def reconnect(self):
        self.close()
        THREAD_LOCAL.socket = self.get_socket()

    @staticmethod
    def close():
        if THREAD_LOCAL.socket is not None:
            try:
                THREAD_LOCAL.socket.close()
            except socket.error:
                pass

    def get_socket(self):
        _socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        _socket.connect(self.ipc_path)
        # Tell the socket not to block on reads.
        _socket.settimeout(2)
        return _socket

    @staticmethod
    def send_request(request):
        THREAD_LOCAL.socket.sendall(request)

    @staticmethod
    def read_response():
        response_raw = ""
        while True:
            response_raw += to_text(THREAD_LOCAL.socket.recv(4096))
            trimmed = response_raw.strip()
            if trimmed and trimmed[-1] == '}':
                logging.debug("ipc response: %s" % response_raw)
                try:
                    return json.loads(trimmed)
                except JSONDecodeError:
                    continue

    def _call(self, method, params=None, _id=1):
        params = params or []
        data = {
            'jsonrpc': '2.0',
            'method': method,
            'params': params,
            'id': _id,
        }
        logging.debug("ipc send: %s" % json.dumps(data))
        request = to_bytes(json.dumps(data))

        self.connect_if_not_yet()

        for _ in range(3):
            try:
                self.send_request(request)
                response = self.read_response()
                break
            except socket.timeout or OSError:
                self.reconnect()
        else:
            raise ValueError("No JSON returned by socket")

        if "error" in response:
            raise ValueError(response["error"]["message"])

        return response["result"]



import re
from mythril.analysis.ops import get_variable, VarType
from mythril.analysis.report import Issue
import logging


'''
MODULE DESCRIPTION:

Check for invocations of delegatecall(msg.data) in the fallback function.
'''


def execute(statespace):

    issues = []

    for call in statespace.calls:

        if (call.type == "DELEGATECALL"):

            state = call.state
            address = state.get_current_instruction()['address']

            if (call.node.function_name == "fallback"):

                stack = state.mstate.stack
                meminstart = get_variable(stack[-3])

                if meminstart.type == VarType.CONCRETE:

                    if (re.search(r'calldata.*_0', str(state.mstate.memory[meminstart.val]))):

                        issue = Issue(call.node.contract_name, call.node.function_name, address, "Call data forwarded with delegatecall()", "Informational")

                        issue.description = \
                            "This contract forwards its call data via DELEGATECALL in its fallback function. " \
                            "This means that any function in the called contract can be executed. Note that the callee contract will have access to the storage of the calling contract.\n"   

                        if (call.to.type == VarType.CONCRETE):
                            issue.description += ("DELEGATECALL target: " + hex(call.to.val))
                        else:
                            issue.description += "DELEGATECALL target: " + str(call.to)

                        issues.append(issue)

                if (call.to.type == VarType.SYMBOLIC):

                    issue = Issue(call.node.contract_name, call.node.function_name, address, call.type + " to a user-supplied address")

                    if ("calldata" in str(call.to)):

                        issue.description = \
                            "This contract delegates execution to a contract address obtained from calldata. "

                    else:
                        m = re.search(r'storage_([a-z0-9_&^]+)', str(call.to))

                        if (m):
                            idx = m.group(1)

                        func = statespace.find_storage_write(state.environment.active_account.address, idx)

                        if (func):
                            issue.description = "This contract delegates execution to a contract address in storage slot " + str(idx) + ". This storage slot can be written to by calling the function `" + func + "`. "

                        else:
                            logging.debug("[DELEGATECALL] No storage writes to index " + str(idx))

                        issue.description += "Be aware that the called contract gets unrestricted access to this contract's state."
                        issues.append(issue)

    return issues



import plyvel
from ethereum.db import BaseDB
from ethereum import utils

class ETH_DB(BaseDB):
    '''
    adopts pythereum BaseDB using plyvel
    '''

    def __init__(self, path):
        self.db = plyvel.DB(path)

    def get(self, key):
        '''
        gets value for key
        '''
        return self.db.get(key)

    def put(self, key, value):
        '''
        puts value for key
        '''
        self.db.put(key, value)